@@ -5,7 +5,7 @@
 {-# LANGUAGE DeriveDataTypeable #-}
 #endif
 
-#if defined(__GLASGOW_HASKELL__) && __GLASGOW_HASKELL__ >= 702
+#if __GLASGOW_HASKELL__ >= 702
 #define LANGUAGE_DefaultSignatures
 {-# LANGUAGE DefaultSignatures #-}
 #if defined(MIN_VERSION_hashable) || __GLASGOW_HASKELL__ >= 708
@@ -15,15 +15,26 @@
 #endif
 #endif
 
-#if defined(__GLASGOW_HASKELL__) && __GLASGOW_HASKELL__ >= 704
+#if __GLASGOW_HASKELL__ >= 704
 #define LANGUAGE_DeriveGeneric
 {-# LANGUAGE DeriveGeneric #-}
+{-# LANGUAGE TypeOperators #-}
+{-# LANGUAGE FlexibleContexts #-}
+#endif
+
+#if __GLASGOW_HASKELL__ >= 708
+#define USE_COERCE
+{-# LANGUAGE ScopedTypeVariables #-}
+#endif
+
+#ifndef MIN_VERSION_base
+#define MIN_VERSION_base(x,y,z) 1
 #endif
 
 -----------------------------------------------------------------------------
 -- |
 -- Module      :  Data.Semigroup
--- Copyright   :  (C) 2011-2014 Edward Kmett
+-- Copyright   :  (C) 2011-2015 Edward Kmett
 -- License     :  BSD-style (see the file LICENSE)
 --
 -- Maintainer  :  Edward Kmett <ekmett@gmail.com>
@@ -68,18 +79,30 @@ module Data.Semigroup (
   -- * Difference lists of a semigroup
   , diff
   , cycle1
+  -- * ArgMin, ArgMax
+  , Arg(..)
+  , ArgMin
+  , ArgMax
   ) where
 
 import Prelude hiding (foldr1)
-import Data.Monoid (Monoid(..),Dual(..),Endo(..),All(..),Any(..),Sum(..),Product(..),Endo(..))
+
+#if MIN_VERSION_base(4,8,0)
+import Data.Bifunctor
+#else
+import Data.Monoid (Monoid(..))
+import Data.Foldable
+import Data.Traversable
+#endif
+
+import Data.Monoid (Dual(..),Endo(..),All(..),Any(..),Sum(..),Product(..))
+
 import Control.Applicative
 import Control.Monad
 import Control.Monad.Fix
 import qualified Data.Monoid as Monoid
-import Data.Foldable
-import Data.Traversable
 import Data.List.NonEmpty
-import Numeric.Natural.Internal
+import Numeric.Natural
 
 #ifdef MIN_VERSION_deepseq
 import Control.DeepSeq (NFData(..))
@@ -96,11 +119,22 @@ import Data.IntMap (IntMap)
 #ifdef MIN_VERSION_bytestring
 import Data.ByteString as Strict
 import Data.ByteString.Lazy as Lazy
+
+# if MIN_VERSION_bytestring(0,10,2)
+import qualified Data.ByteString.Builder as ByteString
+# elif MIN_VERSION_bytestring(0,10,0)
+import qualified Data.ByteString.Lazy.Builder as ByteString
+# endif
+
+# if MIN_VERSION_bytestring(0,10,4)
+import Data.ByteString.Short
+# endif
 #endif
 
 #ifdef MIN_VERSION_text
 import qualified Data.Text as Strict
 import qualified Data.Text.Lazy as Lazy
+import qualified Data.Text.Lazy.Builder as Text
 #endif
 
 #ifdef MIN_VERSION_hashable
@@ -120,6 +154,10 @@ import Data.Data
 import GHC.Generics
 #endif
 
+#ifdef USE_COERCE
+import Data.Coerce
+#endif
+
 infixr 6 <>
 
 class Semigroup a where
@@ -160,17 +198,17 @@ class Semigroup a where
   --
   -- See also 'timesN'.
 
-  times1p :: Whole n => n -> a -> a
+  times1p :: Natural -> a -> a
   times1p y0 x0 = f x0 (1 Prelude.+ y0)
     where
       f x y
         | even y = f (x <> x) (y `quot` 2)
         | y == 1 = x
-        | otherwise = g (x <> x) (unsafePred y  `quot` 2) x
+        | otherwise = g (x <> x) (pred y  `quot` 2) x
       g x y z
         | even y = g (x <> x) (y `quot` 2) z
         | y == 1 = x <> z
-        | otherwise = g (x <> x) (unsafePred y `quot` 2) (x <> z)
+        | otherwise = g (x <> x) (pred y `quot` 2) (x <> z)
   {-# INLINE times1p #-}
 
 -- | A generalization of 'Data.List.cycle' to an arbitrary 'Semigroup'.
@@ -228,24 +266,48 @@ instance Semigroup a => Semigroup (Dual a) where
   times1p n (Dual a) = Dual (times1p n a)
 
 instance Semigroup (Endo a) where
+#ifdef USE_COERCE
+  (<>) = coerce ((.) :: (a -> a) -> (a -> a) -> (a -> a))
+#else
   Endo f <> Endo g = Endo (f . g)
+#endif
 
 instance Semigroup All where
+#ifdef USE_COERCE
+  (<>) = coerce (&&)
+#else
   All a <> All b = All (a && b)
+#endif
   times1p _ a = a
 
 instance Semigroup Any where
+#ifdef USE_COERCE
+  (<>) = coerce (||)
+#else
   Any a <> Any b = Any (a || b)
+#endif
   times1p _ a = a
 
 instance Num a => Semigroup (Sum a) where
+#ifdef USE_COERCE
+  (<>) = coerce ((+) :: a -> a -> a)
+#else
   Sum a <> Sum b = Sum (a + b)
+#endif
 
 instance Num a => Semigroup (Product a) where
+#ifdef USE_COERCE
+  (<>) = coerce ((*) :: a -> a -> a)
+#else
   Product a <> Product b = Product (a * b)
+#endif
 
 instance Semigroup a => Semigroup (Const a b) where
+#ifdef USE_COERCE
+  (<>) = coerce ((<>) :: a -> a -> a)
+#else
   Const a <> Const b = Const (a <> b)
+#endif
 
 #if MIN_VERSION_base(3,0,0)
 instance Semigroup (Monoid.First a) where
@@ -262,6 +324,7 @@ instance Semigroup (Monoid.Last a) where
 instance Semigroup (NonEmpty a) where
   (a :| as) <> ~(b :| bs) = a :| (as ++ b : bs)
 
+
 newtype Min a = Min { getMin :: a } deriving
   ( Eq, Ord, Show, Read
 #ifdef LANGUAGE_DeriveDataTypeable
@@ -269,6 +332,9 @@ newtype Min a = Min { getMin :: a } deriving
 #endif
 #ifdef LANGUAGE_DeriveGeneric
   , Generic
+#if __GLASGOW_HASKELL__ >= 706
+  , Generic1
+#endif
 #endif
   )
 
@@ -296,7 +362,11 @@ instance Hashable a => Hashable (Min a) where
 #endif
 
 instance Ord a => Semigroup (Min a) where
+#ifdef USE_COERCE
+  (<>) = coerce (min :: a -> a -> a)
+#else
   Min a <> Min b = Min (a `min` b)
+#endif
   times1p _ a = a
 
 instance (Ord a, Bounded a) => Monoid (Min a) where
@@ -338,6 +408,9 @@ newtype Max a = Max { getMax :: a } deriving
 #endif
 #ifdef LANGUAGE_DeriveGeneric
   , Generic
+#if __GLASGOW_HASKELL__ >= 706
+  , Generic1
+#endif
 #endif
   )
 
@@ -365,7 +438,11 @@ instance Hashable a => Hashable (Max a) where
 #endif
 
 instance Ord a => Semigroup (Max a) where
+#ifdef USE_COERCE
+  (<>) = coerce (max :: a -> a -> a)
+#else
   Max a <> Max b = Max (a `max` b)
+#endif
   times1p _ a = a
 
 instance (Ord a, Bounded a) => Monoid (Max a) where
@@ -400,6 +477,64 @@ instance NFData a => NFData (Max a) where
   rnf (Max a) = rnf a
 #endif
 
+-- | 'Arg' isn't itself a 'Semigroup' in its own right, but it can be placed inside 'Min' and 'Max'
+-- to compute an arg min or arg max.
+data Arg a b = Arg a b deriving
+  ( Show, Read
+#ifdef LANGUAGE_DeriveDataTypeable
+  , Data, Typeable
+#endif
+#ifdef LANGUAGE_DeriveGeneric
+  , Generic
+#if __GLASGOW_HASKELL__ >= 706
+  , Generic1
+#endif
+#endif
+  )
+
+type ArgMin a b = Min (Arg a b)
+type ArgMax a b = Max (Arg a b)
+
+instance Functor (Arg a) where
+  fmap f (Arg x a) = Arg x (f a)
+
+instance Foldable (Arg a) where
+  foldMap f (Arg _ a) = f a
+
+instance Traversable (Arg a) where
+  traverse f (Arg x a) = Arg x <$> f a
+
+instance Eq a => Eq (Arg a b) where
+  Arg a _ == Arg b _ = a == b
+
+instance Ord a => Ord (Arg a b) where
+  Arg a _ `compare` Arg b _ = compare a b
+  min x@(Arg a _) y@(Arg b _)
+    | a <= b    = x
+    | otherwise = y
+  max x@(Arg a _) y@(Arg b _)
+    | a >= b    = x
+    | otherwise = y
+
+#ifdef MIN_VERSION_deepseq
+instance (NFData a, NFData b) => NFData (Arg a b) where
+  rnf (Arg a b) = rnf a `seq` rnf b `seq` ()
+#endif
+
+#ifdef MIN_VERSION_hashable
+instance (Hashable a, Hashable b) => Hashable (Arg a b) where
+#if MIN_VERSION_hashable(1,2,0)
+  hashWithSalt p (Arg a b) = hashWithSalt p a `hashWithSalt` b
+#else
+  hash (Arg a b) = hashWithSalt (hash a) b
+#endif
+#endif
+
+#if MIN_VERSION_base(4,8,0)
+instance Bifunctor Arg where
+  bimap f g (Arg a b) = Arg (f a) (g b)
+#endif
+
 -- | Use @'Option' ('First' a)@ to get the behavior of 'Data.Monoid.First' from @Data.Monoid@.
 newtype First a = First { getFirst :: a } deriving
   ( Eq, Ord, Show, Read
@@ -409,6 +544,9 @@ newtype First a = First { getFirst :: a } deriving
 #endif
 #ifdef LANGUAGE_DeriveGeneric
   , Generic
+#if __GLASGOW_HASKELL__ >= 706
+  , Generic1
+#endif
 #endif
   )
 
@@ -475,6 +613,9 @@ newtype Last a = Last { getLast :: a } deriving
 #endif
 #ifdef LANGUAGE_DeriveGeneric
   , Generic
+#if __GLASGOW_HASKELL__ >= 706
+  , Generic1
+#endif
 #endif
   )
 
@@ -542,6 +683,16 @@ instance Semigroup Strict.ByteString where
 
 instance Semigroup Lazy.ByteString where
   (<>) = mappend
+
+# if MIN_VERSION_bytestring(0,10,0)
+instance Semigroup ByteString.Builder where
+  (<>) = mappend
+# endif
+
+# if MIN_VERSION_bytestring(0,10,4)
+instance Semigroup ShortByteString where
+  (<>) = mappend
+# endif
 #endif
 
 #ifdef MIN_VERSION_text
@@ -550,6 +701,9 @@ instance Semigroup Strict.Text where
 
 instance Semigroup Lazy.Text where
   (<>) = mappend
+
+instance Semigroup Text.Builder where
+  (<>) = mappend
 #endif
 
 #ifdef MIN_VERSION_unordered_containers
@@ -570,6 +724,9 @@ newtype WrappedMonoid m = WrapMonoid
 #endif
 #ifdef LANGUAGE_DeriveGeneric
   , Generic
+#if __GLASGOW_HASKELL__ >= 706
+  , Generic1
+#endif
 #endif
   )
 
@@ -583,11 +740,15 @@ instance Hashable a => Hashable (WrappedMonoid a) where
 #endif
 
 instance Monoid m => Semigroup (WrappedMonoid m) where
+#ifdef USE_COERCE
+  (<>) = coerce (mappend :: m -> m -> m)
+#else
   WrapMonoid a <> WrapMonoid b = WrapMonoid (a `mappend` b)
+#endif
 
 instance Monoid m => Monoid (WrappedMonoid m) where
   mempty = WrapMonoid mempty
-  WrapMonoid a `mappend` WrapMonoid b = WrapMonoid (a `mappend` b)
+  mappend = (<>)
 
 instance Bounded a => Bounded (WrappedMonoid a) where
   minBound = WrapMonoid minBound
@@ -613,9 +774,9 @@ instance NFData m => NFData (WrappedMonoid m) where
 -- > timesN n a = a <> a <> ... <> a  -- using <> (n-1) times
 --
 -- Implemented using 'times1p'.
-timesN :: (Whole n, Monoid a) => n -> a -> a
+timesN :: Monoid a => Natural -> a -> a
 timesN n x | n == 0    = mempty
-           | otherwise = unwrapMonoid . times1p (unsafePred n) . WrapMonoid $ x
+           | otherwise = unwrapMonoid . times1p (pred n) . WrapMonoid $ x
 {-# INLINE timesN #-}
 
 
@@ -631,6 +792,9 @@ newtype Option a = Option
 #endif
 #ifdef LANGUAGE_DeriveGeneric
   , Generic
+#if __GLASGOW_HASKELL__ >= 706
+  , Generic1
+#endif
 #endif
   )
 
@@ -689,11 +853,15 @@ option :: b -> (a -> b) -> Option a -> b
 option n j (Option m) = maybe n j m
 
 instance Semigroup a => Semigroup (Option a) where
+#ifdef USE_COERCE
+  (<>) = coerce ((<>) :: Maybe a -> Maybe a -> Maybe a)
+#else
   Option a <> Option b = Option (a <> b)
+#endif
 
 instance Semigroup a => Monoid (Option a) where
   mempty = Option Nothing
-  Option a `mappend` Option b = Option (a <> b)
+  mappend = (<>)
 
 -- | This lets you use a difference list of a 'Semigroup' as a 'Monoid'.
 diff :: Semigroup m => m -> Endo m
